"""This module provides core functionalities of gntplib.

gntplib is a Growl Notification Transport Protocol (GNTP_) client library in
Python.

.. _GNTP: http://www.growlforwindows.com/gfw/help/gntp.aspx
"""

import hashlib
import io
import re
import socket


__version__ = '0.2'
__all__ = ['notify', 'Notifier', 'RawIcon', 'SocketCallback', 'Event']


SUPPORTED_VERSIONS = ['1.0']
MAX_MESSAGE_SIZE = 4096
MAX_LINE_SIZE = 1024
LINE_DELIMITER = '\r\n'
SECTION_DELIMITER = SECTION_BODY_START = SECTION_BODY_END = '\r\n'
MESSAGE_DELIMITER = '\r\n\r\n'
RESPONSE_INFORMATION_LINE_RE = re.compile(
    'GNTP/([^ ]+) (-OK|-ERROR|-CALLBACK) NONE')


def notify(app_name, event_name, title, text=''):
    """Register notifier and notify event at a time.

    :param app_name: the name of the application.
    :param event_name: the name of the event.
    :param title: the title of the notification.
    :param text: the text of the notification.  Defaults to ``''``.
    """
    notifier = Notifier(app_name, coerce_to_events([event_name]))
    notifier.register()
    notifier.notify(event_name, title, text)


class Notifier(object):
    """Notifier of Growl Notification Transport Protocol (GNTP).

    Currently this class supports ``REGISTER`` and ``NOTIFY`` requests.  They
    are sent by :meth:`register()` and :meth:`notify()` methods respectively.
    These methods can accept the optional final callback as `callback` keyword
    argument, which run after closing the connection with the GNTP server.

    `event_defs` is list of ``str``, ``unicode``, double or :class:`Event`
    instance.  It is converted to list of :class:`Event` instance as follows:
    ``str`` or ``unicode`` item becomes value of the `name` attribute of
    :class:`Event` instance, whose other attributes are defaults.  Double item
    is expanded to (`name`, `enabled`) tuple, and those values are passed to
    :class:`Event` constructor.  :class:`Event` instance item is used directly.

    Optional keyword arguments are passed to the `gntp_client_class`
    constructor .

    :param name: the name of the application.
    :param event_defs: the definitions of the events.
    :param icon: :class:`Icon` instance or string for the application icon.
                 Defaults to `None`.
    :param gntp_client_class: GNTP client class.  If it is `None`,
                              :class:`GNTPClient` is used.  Defaults to `None`.

    .. note:: In Growl 1.3.3, `icon` of string does not work.
    """

    def __init__(self, name, event_defs, icon=None, gntp_client_class=None,
                 **kwargs):
        self.name = name
        self.icon = icon
        self.events = coerce_to_events(event_defs)
        if gntp_client_class is None:
            gntp_client_class = GNTPClient
        self.gntp_client = gntp_client_class(**kwargs)

    def register(self, _message_callback=None, callback=None):
        """Register this notifier to GNTP server.

        :param _message_callback: only for use in unit testing.
        :param callback: the callback run after closing the connection with
                         the GNTP server.
        """
        request = RegisterRequest(self.name, self.icon, self.events)
        self.gntp_client.register(request, _message_callback, callback)

    def notify(self, name, title, text='', id_=None, sticky=False,
               priority=0, icon=None, coalescing_id=None, callback=None,
               _message_callback=None, _callback_message_callback=None,
               gntp_callback=None, **socket_callback_options):
        """Notify notification to GNTP server.

        :param name: the name of the event.
        :param title: the title of the notification.
        :param text: the text of the notification.  Defaults to `''`.
        :param id_: the unique ID for the notification.  Defaults to `None`.
        :param sticky: if the notification should remain displayed until
                       dismissed by the user.  Defaults to `False`.
        :param priority: the display hint for the receiver which may be
                         ignored.  A higher number indicates a higher priority.
                         Valid values are between -2 and 2, defaults to `0`.
        :param icon: :class:`Icon` instance or string to display with the
                     notification.  Defaults to `None`.
        :param coalescing_id: if present, should contain the value of the
                              `id_` of a previously-sent notification.
                              This serves as a hint to the notification system
                              that this notification should replace/update the
                              matching previous notification.  The notification
                              system may ignore this hint.  Defaults to `None`.
        :param callback: the callback run after closing the connection with
                         the GNTP server.
        :param _message_callback: only for use in unit testing.
        :param _callback_message_callback: only for use in unit testing.
        :param gntp_callback: url string for url callback or
                              :class:`SocketCallback` for socket callback.
                              Defaults to `None`.
        :param socket_callback_options: the keyword arguments to used to
                                        instantiating :class:`SocketCallback`
                                        for socket callback.  About acceptable
                                        keyword arguments,
                                        see :func:`coerce_to_callback()`.

        .. note:: In Growl 1.3.3, `icon` of string does not work.

        """
        notification = Notification(name, title, text, id_=id_, sticky=sticky,
                                    priority=priority, icon=icon,
                                    coalescing_id=coalescing_id,
                                    gntp_callback=gntp_callback,
                                    **socket_callback_options)
        request = NotifyRequest(self.name, notification)
        self.gntp_client.notify(request, _message_callback, callback,
                                _callback_message_callback,
                                notification.socket_callback)


class GNTPError(Exception):
    """Base class for GNTP exception."""


class Event(object):
    """Represent notification type.

    :param name: the name of the notification.
    :param display_name: the display name of the notification, which is shown
                         at the Applications tab of the Growl Preferences.
                         Defaults to `None`.
    :param enabled: indicates if the notification should be enabled by
                    default.  Defaults to `True`.
    :param icon: :class:`Icon` instance or string, the default icon to use
                 for notifications of this type.  Defaults to `None`.

    .. note:: In Growl 1.3.3, `icon` does not work.
    """

    def __init__(self, name, display_name=None, enabled=True, icon=None):
        self.name = name
        self.display_name = display_name
        self.enabled = enabled
        self.icon = coerce_to_icon(icon)


class Notification(object):
    """Represent notification."""

    def __init__(self, name, title, text='', id_=None, sticky=None,
                 priority=None, icon=None, coalescing_id=None,
                 gntp_callback=None, **socket_callback_options):
        self.name = name
        self.title = title
        self.text = text
        self.id_ = id_
        self.sticky = sticky
        self.priority = priority
        self.icon = coerce_to_icon(icon)
        self.coalescing_id = coalescing_id
        self.callback = coerce_to_callback(gntp_callback,
                                           **socket_callback_options)

    @property
    def socket_callback(self):
        if isinstance(self.callback, SocketCallback):
            return self.callback


class BaseRequest(object):
    """Abstract base class for GNTP request."""
    #: Request message type.  Subclasses must override this attribute.
    #: Currently ``REGISTER`` and ``NOTIFY`` are supported.
    message_type = None

    @property
    def message(self):
        """Return utf-8 encoded request message."""
        return 'GNTP/1.0 %s NONE\r\n%s\r\n' % (
            self.message_type, self.message_body)

    @property
    def message_body(self):
        """Return utf-8 encoded message body."""
        writer = io.BytesIO()
        header_writer = HeaderWriter(writer)
        section_writer = SectionWriter(writer, header_writer)
        self._write_into(header_writer)
        self._write_into(section_writer)
        return writer.getvalue()

    def _write_into(self, writer):
        """Subclasses must override this method to serialize their message
        bodies.
        """
        raise NotImplementedError


class RegisterRequest(BaseRequest):
    """Represent ``REGISTER`` request.

    :param app_name: the name of the application.
    :param app_icon: :class:`Icon` instance or string for the application.
    :param events: the list of the events.

    .. note:: In Growl 1.3.3, `app_icon` of string does not work.
    """
    message_type = 'REGISTER'

    def __init__(self, app_name, app_icon, events):
        self.app_name = app_name
        self.app_icon = coerce_to_icon(app_icon)
        self.events = events

    def _write_into(self, writer):
        writer.write_register_request(self)


class NotifyRequest(BaseRequest):
    """Represent ``NOTIFY`` request.

    :param app_name: the name of the application.
    :param notification: :class:`Notification` instance.
    """
    message_type = 'NOTIFY'

    def __init__(self, app_name, notification):
        self.app_name = app_name
        self.notification = notification

    def _write_into(self, writer):
        writer.write_notify_request(self)


class Response(object):
    """Base class for GNTP response."""

    def __init__(self, message_type, headers):
        self.message_type = message_type
        self.headers = headers


class GNTPConnectionCallbacksMixin(object):
    """Defines callbacks run by the GNTP connections."""

    def on_register_message(self, message):
        """Callback for ``-OK`` response of ``REGISTER`` request.

        :param message: string of response terminated by
                        `'\\\\r\\\\n\\\\r\\\\n\\\\r\\\\n'`.
        """
        try:
            response = parse_response(message, '-OK')
        finally:
            self.close()
        if self.final_callback is not None:
            self.final_callback(response)

    def on_notify_message(self, message):
        """Callback for ``-OK`` response of ``NOTIFY`` request.

        :param message: string of response terminated by
                        `'\\\\r\\\\n\\\\r\\\\n\\\\r\\\\n'`.
        """
        try:
            response = parse_response(message, '-OK')
            if self.socket_callback is not None:
                self.read_message(self.callback_message_callback)
        finally:
            if self.socket_callback is None:
                self.close()
        if self.socket_callback is None and self.final_callback is not None:
            self.final_callback(response)

    def on_callback_message(self, message):
        """Callback for ``-CALLBACK`` response.

        :param message: string of response terminated by
                        `'\\\\r\\\\n\\\\r\\\\n\\\\r\\\\n'`.
        """
        try:
            callback_response = parse_response(message, '-CALLBACK')
            response = self.socket_callback(callback_response)
        finally:
            self.close()
        if self.final_callback is not None:
            self.final_callback(response)


class GNTPConnection(GNTPConnectionCallbacksMixin):
    """Represent the connection with the GNTP server."""

    def __init__(self, address, timeout, final_callback,
                 callback_message_callback=None, socket_callback=None):
        self.sock = socket.create_connection(address, timeout=timeout)
        self.final_callback = final_callback
        self.callback_message_callback = (callback_message_callback or
                                          self.on_callback_message)
        self.socket_callback = socket_callback

    def write_message(self, request):
        """Send the request message to the GNTP server."""
        self.sock.send(request.message)

    def read_message(self, callback):
        """Read a message from opened socket and run callback with it."""
        message = next(generate_messages(self.sock))
        callback(message)

    def close(self):
        """Close the connection."""
        self.sock.close()
        self.sock = None


class GNTPClient(object):
    """Tiny GNTP client

    :param host: host of GNTP server.  Defaults to ``'localhost'``.
    :param port: port of GNTP server.  Defaults to ``23053``.
    :param timeout: timeout.  Defaults to ``10``.
    """

    connection_class = GNTPConnection

    def __init__(self, host='localhost', port=23053, timeout=10):
        self.address = (host, port)
        self.timeout = timeout

    def register(self, request, message_callback, callback):
        """Open connection and handle register request.

        :param message_callback: the callback run with the response message of
                                 the ``REGISTER`` request.  This callback must
                                 close the connection.
        :param callback: the final callback run after closing connection.
        """
        conn = self._connect(callback)
        conn.write_message(request)
        conn.read_message(message_callback or conn.on_register_message)

    def notify(self, request, message_callback, callback,
               callback_message_callback, socket_callback):
        """Open connection and handle notify request.

        :param message_callback: the callback run with the response message of
                                 the ``NOTIFY`` request.  This callback must
                                 close the connection if the ``-CALLBACK``
                                 response will not follow..
        :param callback: the final callback run after closing connection.
        :param callback_message_callback: the callback run with the
                                          ``-CALLBACK`` response message.
                                          If not `None`, must close the
                                          connection.
        :param socket_callback: :class:`SocketCallback` instance called by
                                `callback_message_callback`.
        """
        conn = self._connect(
            callback,
            callback_message_callback=callback_message_callback,
            socket_callback=socket_callback)
        conn.write_message(request)
        conn.read_message(message_callback or conn.on_notify_message)

    def _connect(self, final_callback, **kwargs):
        """Connect to GNTP server and return the connection."""
        return self.connection_class(self.address, self.timeout,
                                     final_callback, **kwargs)


def generate_messages(sock, size=1024):
    """Generate messages from opened socket."""
    buf = ''
    while True:
        buf += sock.recv(size)
        if not buf:
            break
        # or .find(LINE_DELIMITER + MESSAGE_DELIMITER)?
        pos = buf.find(MESSAGE_DELIMITER)
        if ((pos < 0 and len(buf) >= MAX_MESSAGE_SIZE) or
            (pos > MAX_MESSAGE_SIZE - 4)):
            raise GNTPError('too large message: %r' % buf)
        elif pos > 0:
            pos += 4
            yield buf[:pos]
            buf = buf[pos:]


def parse_response(message, expected_message_type=None):
    """Parse response and return response object."""
    try:
        lines = message.split(LINE_DELIMITER)
        _, message_type = parse_information_line(lines.pop(0))
        if (expected_message_type is not None and
            expected_message_type != message_type):
            raise GNTPError('%s is not expected message type %s' % (
                    message_type, expected_message_type))

        headers = dict([s.strip() for s in line.split(':', 1)]
                       for line in lines if line)
        if message_type == '-ERROR':
            raise GNTPError('%s: %s' % (headers['Error-Code'],
                                        headers['Error-Description']))
        return Response(message_type, headers)
    except (ValueError, GNTPError) as why:
        raise GNTPError(why.args[0], 'original message: %s' % message)


def parse_information_line(line):
    """Parse information line and return tuple (`<version>`,
    `<messagetype>`)."""
    matched = RESPONSE_INFORMATION_LINE_RE.match(line)
    if matched is None:
        raise GNTPError('invalid information line: %r' % line)
    version, message_type = matched.groups()
    if version not in SUPPORTED_VERSIONS:
        raise GNTPError('version %r is not supported' % version)
    return version, message_type


def coerce_to_events(items):
    """Coerce list of event definitions to list of :class:`Event` instances."""
    results = []
    for item in items:
        if isinstance(item, basestring):
            results.append(Event(item, enabled=True))
        elif isinstance(item, tuple):
            name, enabled = item
            results.append(Event(name, enabled=enabled))
        elif isinstance(item, Event):
            results.append(item)
    return results


class BaseIcon(object):
    """Abstract base class for icon."""

    def __init__(self, data):
        self.data = data

    def resource(self):
        """Used to write header of this instance into the writer."""
        raise NotImplementedError

    def write_into(self, writer):
        """Used to write section of this instance into the writer."""
        raise NotImplementedError


class RawIcon(BaseIcon):
    """Class for icon.

    :param data: binary data of image.
    """

    def __init__(self, data):
        BaseIcon.__init__(self, data)
        self._unique_id = None

    def unique_id(self):
        """Return the value of <uniqueid> data type."""
        if self.data is not None and self._unique_id is None:
            self._unique_id = hashlib.md5(self.data).hexdigest()
        return self._unique_id

    def resource(self):
        """Return the resource string used in request headers."""
        if self.data is not None:
            return 'x-growl-resource://%s' % self.unique_id()

    def write_into(self, writer):
        if self.data is not None:
            writer.write_raw_icon(self)


class URLIcon(BaseIcon):

    def __init__(self, url):
        BaseIcon.__init__(self, url)

    def resource(self):
        return self.data

    def write_into(self, writer):
        pass


def coerce_to_icon(obj):
    if obj is None or isinstance(obj, BaseIcon):
        return obj
    if isinstance(obj, basestring):
        return URLIcon(obj)


class SocketCallback(object):
    """Base class for socket callback.

    :param context: value of ``Notification-Callback-Context``.
    :param context-type: value of ``Notification-Callback-Context-Type``.
                         Defaults to ``'text/plain; charset=utf-8'``.
    :param on_click: the function run at ``CLICKED`` callback result.
    :param on_close: the function run at ``CLOSED`` callback result.
    :param on_timeout: the function run at ``CLOSED`` callback result.
    """

    def __init__(self, context='', context_type='text/plain; charset=utf-8',
                 on_click=None, on_close=None, on_timeout=None):
        self.context = context
        self.context_type = context_type
        self.on_click_callback = on_click
        self.on_close_callback = on_close
        self.on_timeout_callback = on_timeout

    def on_click(self, response):
        """Run ``CLICKED`` event callback."""
        if self.on_click_callback is not None:
            return self.on_click_callback(response)

    def on_close(self, response):
        """Run ``CLOSED`` event callback."""
        if self.on_close_callback is not None:
            return self.on_close_callback(response)

    def on_timeout(self, response):
        """Run ``TIMEDOUT`` event callback."""
        if self.on_timeout_callback is not None:
            return self.on_timeout_callback(response)

    def __call__(self, response):
        """This is the callback.  Delegate to ``on_`` methods depending on
        ``Notification-Callback-Result`` value.

        :param response: :class:`Response` instance.

        .. note:: TIMEDOUT callback does not occur in my Growl 1.3.3.
        """
        callback_result = response.headers['Notification-Callback-Result']
        delegate_map = {
            'CLICKED': self.on_click, 'CLICK': self.on_click,
            'CLOSED': self.on_close, 'CLOSE': self.on_close,
            'TIMEDOUT': self.on_timeout, 'TIMEOUT': self.on_timeout,
            }
        return delegate_map[callback_result](response)

    def write_into(self, writer):
        writer.write_socket_callback(self)


class URLCallback(object):

    def __init__(self, url):
        self.url = url

    def write_into(self, writer):
        writer.write_url_callback(self)


def coerce_to_callback(gntp_callback=None, **socket_callback_options):
    """Return :class:`URLCallback` instance for url callback or
    :class:`SocketCallback` instance for socket callback.

    If `gntp_callback` is not `None`, `socket_callback_options` must be empty.
    If `gntp_callback` is a instance of ``basestring``, then a instance of
    :class:`URLCallback` is returned.  Otherwise, `gntp_callback` is returned
    directly.

    If `gntp_callback` is `None` and `socket_callback_options` is not empty,
    new instance of :class:`SocketCallback` is created from given keyword
    arguments and it is returned.  Acceptable keyword arguments are same as
    constructor's of :class:`SocketCallback`.
    """
    if gntp_callback is not None:
        if socket_callback_options:
            raise GNTPError('If gntp_callback is not None,'
                            ' socket_callback_options must be empty')
        if isinstance(gntp_callback, basestring):
            return URLCallback(gntp_callback)
        else:
            return gntp_callback
    if socket_callback_options:
        return SocketCallback(**socket_callback_options)


class HeaderWriter(object):

    def __init__(self, writer):
        self.writer = writer

    def write_register_request(self, request):
        self.write('Application-Name', request.app_name)
        self.write_icon(request.app_icon, 'Application-Icon')
        self.write('Notifications-Count', len(request.events))
        for event in request.events:
            self.writer.write(LINE_DELIMITER)
            self.write_event(event)

    def write_notify_request(self, request):
        self.write('Application-Name', request.app_name)
        self.write_notification(request.notification)

    def write_event(self, event):
        self.write('Notification-Name', event.name)
        self.write('Notification-Display-Name', event.display_name)
        self.write('Notification-Enabled', event.enabled)
        self.write_icon(event.icon, 'Notification-Icon')

    def write_notification(self, notification):
        self.write('Notification-Name', notification.name)
        self.write('Notification-ID', notification.id_)
        self.write('Notification-Title', notification.title)
        self.write('Notification-Text', notification.text)
        self.write('Notification-Sticky', notification.sticky)
        self.write('Notification-Priority', notification.priority)
        self.write_icon(notification.icon, 'Notification-Icon')
        self.write('Notification-Coalescing-ID', notification.coalescing_id)
        if notification.callback is not None:
            notification.callback.write_into(self)

    def write_socket_callback(self, callback):
        self.write('Notification-Callback-Context', callback.context)
        self.write('Notification-Callback-Context-Type', callback.context_type)

    def write_url_callback(self, callback):
        self.write('Notification-Callback-Target', callback.url)

    def write_icon(self, icon, name):
        if icon is not None:
            self.write(name, icon.resource())

    def write(self, name, value):
        """Write utf-8 encoded headers into writer.

        `headers` is iterable of (`header name`, `header value`) tuple.
        If `header value` is instance of unicode, it is encoded.
        Otherwise it is converted with ``str`` function.
        """
        if value is not None:
            if isinstance(value, unicode):
                value = value.encode('utf-8')
            elif not isinstance(value, str):
                value = str(value)
            self.writer.write(name)
            self.writer.write(': ')
            self.writer.write(value)
            self.writer.write(LINE_DELIMITER)


class SectionWriter(object):

    def __init__(self, writer, header_writer):
        self.writer = writer
        self.header_writer = header_writer

    def write_register_request(self, request):
        self.write_icon(request.app_icon)
        for event in request.events:
            self.write_event(event)

    def write_notify_request(self, request):
        self.write_notification(request.notification)

    def write_event(self, event):
        self.write_icon(event.icon)

    def write_notification(self, notification):
        self.write_icon(notification.icon)

    def write_raw_icon(self, icon):
        headers = [('Identifier', icon.unique_id()),
                   ('Length', len(icon.data))]
        self.write(headers, icon.data)

    def write_icon(self, icon):
        if icon is not None:
            icon.write_into(self)

    def write(self, headers, body):
        """Write utf-8 encoded sections into writer.

        `headers` is iterable of (`section headers`, `section body`) tuple.
        `section headers` is iterable of (`header name`, `header value`)
        tuple.  `section value` is bytes, which is written directly.
        """
        self.writer.write(SECTION_DELIMITER)
        for name, value in headers:
            self.header_writer.write(name, value)
        if body:
            self.writer.write(SECTION_BODY_START)
            self.writer.write(body)
            self.writer.write(SECTION_BODY_END)
